#|
 Copyright (c) 2008 Daniel Herring.
 Distributed under the Boost Software License, Version 1.0.
 (See accompanying file LICENSE_1_0.txt or copy at
  http://www.boost.org/LICENSE_1_0.txt)
|#

#|
CL's reader macros dispatch on one or two characters, and the subchar in set-dispatch-macro-character is always upcased...  This results in a small namespace ripe for collisions and obfuscated names.

This package claims a single top-level macro and dispatches to full names!!!



The default macro character is #?
- #? is explicitely reserved for the user (CLHS section 2.4.8)
- ? indicates conditional reading

Another option is #@
- these keys are next to each other on the standard US keybord
- they aren't reserved for the user (possible standardization ;)

Example:
#?(read-cond-features
    sbcl sb-ext:*core-pathname*
    ecl si:*help-message*
    t "Nothing special")

is equivalent to

#+sbcl sb-ext:*core-pathname*
#+(and ecl (not sbcl)) si:*help-message*
#-(or sbcl ecl) "Nothing special"

|#

(defpackage :read-macros
  (:use :cl)
  (:export

   ;; the dispatcher
   :dispatch
   :install

   ;; conditional reads
   :read-if
   :read-when
   :read-unless
   :read-cond

   ;; conditional reads based on *features*
   :eval-features
   :read-if-feature
   :read-when-feature
   :read-unless-feature
   :read-cond-features

   ;; stateful reads similar to CPP's #if/#elif/#else/#endif
   :read-block-if
   :read-block-else-if
   :read-block-else
   :read-block-end

   ;; lexical read manipulation
   :read-let

   ;; here documents
   :read-here-string
   ))

(in-package :read-macros)


;;;;
;;;; top-level dispatch
;;;;

(defun dispatch (stream subchar infix-param)
  "Read \"#?(RMAC \" and dispatch to the read-macro RMAC.
Technically, the #? are consumed when the lisp reader calls DISPATCH.
This function then checks the #\( and reads RMAC with *package* bound to
:keyword.  Thus RMAC will be read as a keyword unless it has an explicit
package prefix.  Dispatch then funcalls RMAC, or read-macros::RMAC if it
is a keyword."
  ;; these read macros can do evaluation
  ;; in the future, a whitelist may be a good idea
  (unless *read-eval*
    (error 'reader-error :stream stream))
  
  ;; strip the leading #\(
  (peek-char t stream t nil t)
  (unless (eql (read-char stream t nil t)
               #\( )
    (error 'reader-error :stream stream))

  ;; always read the subfunction's name
  (let ((name
         (let ((*package* (find-package :keyword))
               (*read-suppress* nil))
           (read stream t nil t))))
    (when (keywordp name)
      (let ((sym (find-symbol (symbol-name name)
                              (find-package :read-macros))))
        (unless sym
          (error "Could not find reader ~A:~A" :read-macros name))
        (setf name sym)))
    ;; dispatch
    (funcall name stream subchar infix-param)))

;; note: this needs to be called after any changes to DISPATCH
(defun install (&key (subchar #\?) (readtable *readtable*))
  "Install read-macros:dispatch to the readtable.
Defaults to install #? in *readtable*.
Returns the previous read macro for this readtable entry."
  (let ((old (get-dispatch-macro-character #\# subchar readtable)))
    (set-dispatch-macro-character #\# subchar #'dispatch readtable)
    old))

(unless (find :delay-read-macros-install *features*)
  (install))


;;;;
;;;; Utilities
;;;;

;; Odd; I couldn't find anything like cl:eval-features...
(defun eval-features (form)
  "Evaluate FORM according to CLHS 24.1.2.1"
  (cond
    ((not form)
     (error "null features form"))
    ((listp form)
     (ecase (car form)
       (:and
        (let ((result t))
          (dolist (subform (cdr form))
            (setf result (and result (eval-features subform))))
          result))
       (:or 
        (let ((result nil))
          (dolist (subform (cdr form))
            (setf result (or result (eval-features subform))))
          result))
       (:not
        (destructuring-bind (not feature-conditional) form
          (declare (ignore not))
          (not (eval-features feature-conditional))))))
    (t (if (find form *features*)
           t
           nil))))

(defun %read-if (stream
                 subchar
                 infix-param
                 feature-test ; Is test-form features or normal code?
                 pass-result ; Is the desired result t or nil?
                 has-else ; Is there an else clause?
                 )
  (declare (ignore subchar))
  "Common implementation for read-if*, read-when*, and read-unless*."
  (when infix-param
    (error 'reader-error :stream stream))
  (when *read-suppress*
    (return-from %read-if
      (read-delimited-list #\) stream t)))
  (let ((result nil))
    (let* ((form
            (if feature-test
                (let ((*package* (find-package :keyword)))
                  (read stream t nil t))
                (read stream t nil t)))
           (test
            (if feature-test
                (eval-features form)
                (eval form)))
           (passed (eql test pass-result)))
      (when (and has-else
                 (not passed))
        (let ((*read-suppress* t))
          (read stream t nil t)))
      (when (or passed has-else)
        (setf result (multiple-value-list (read stream t nil t))))
      (when (eql passed has-else)
        (let ((*read-suppress* t))
          (read stream t nil t))))

    ;; check there isn't any junk after the form
    ;; and flush the #\) at the same time
    (when (read-delimited-list #\) stream t)
      (error 'reader-error :stream stream))

    (values-list result)))


;;;;
;;;; conditional reads
;;;;

(defun read-if (stream subchar infix-param)
  "#?(read-if test true-form false-form)
 (if (eval (read test))
     (and (read true-form)
          (discard false-form))
     (and (discard true-form)
          (read false-form)))"
  (%read-if stream subchar infix-param
            nil t t))

(defun read-when (stream subchar infix-param)
  "#?(read-when test form)
 (if (eval (read test))
     (read form)
     (discard form))"
  (%read-if stream subchar infix-param
            nil t nil))

(defun read-unless (stream subchar infix-param)
  "#?(read-unless test form)
 (if (eval (read test))
     (discard form)
     (read form))"
  (%read-if stream subchar infix-param
            nil nil nil))

(defun read-cond (stream subchar infix-param)
  (declare (ignore subchar))
  "#?(read-cond test1 form1 test2 form2 ...)
Evaluates each test until one passes and reads the corresponding form.
Everything else is read with *read-suppress* bound to t.
Returns nil if none match"
  (when infix-param
    (error 'reader-error :stream stream))
  (when *read-suppress*
    (read-delimited-list #\) stream t)
    (return-from read-cond nil))
  #| At this point, the stream is " (test-form forms*)*)" |#
  (let ((found nil)
        (result nil))
    ;; loop over "test-form form" (2 reads) until one matches
    (do ((c (peek-char t stream t nil t)
            (peek-char t stream t nil t)))
        ((or found
             (eql c #\) )))
      (let ((*read-suppress*
             (not (eval (read stream t nil t)))))
        (if *read-suppress*
            (read stream t nil t)
            (setf found t
                  result (multiple-value-list (read stream t nil t))))))

    ;; consume any remaining clauses
    (let ((*read-suppress* t))
      (read-delimited-list #\) stream t))

    (values-list result)))


;;;;
;;;; conditional reads based on *features*
;;;;

(defun read-if-feature (stream subchar infix-param)
  "#?(read-if-feature test true-form false-form)
Equivalent to \"#+test true-form #-test false-form\".
Only reads and evaluates test once."
  (%read-if stream subchar infix-param
            t t t))

(defun read-when-feature (stream subchar infix-param)
  "#?(read-when-feature test form)
Equivalent to \"#+test form\"."
  (%read-if stream subchar infix-param
            t t nil))

(defun read-unless-feature (stream subchar infix-param)
  "#?(read-unless-feature test form)
Equivalent to \"#-test form\"."
  (%read-if stream subchar infix-param
            t nil nil))

(defun read-cond-features (stream subchar infix-param)
  (declare (ignore subchar))
  "#?(read-cond-feature test1 form1 test2 form2 ...)
Equivalent to \"#+test1 form1 #+(and test2 (not form1)) form2 ...\""
  (when infix-param
    (error 'reader-error :stream stream))
  (when *read-suppress*
    (return-from read-cond-features
      (read-delimited-list #\) stream t)))
  #| At this point, the stream is " (test-form forms*)*)" |#
  (let ((found nil)
        (result nil))
    ;; loop over "test-form form" (2 reads) until one matches
    (do ((c (peek-char t stream t nil t)
            (peek-char t stream t nil t)))
        ((or found
             (eql c #\) )))
      (let* ((form (let ((*package* (find-package :keyword)))
                     (read stream t nil t)))
             (*read-suppress*
              (not (eval-features form))))
        (if *read-suppress*
            (read stream t nil t)
            (setf found t
                  result (multiple-value-list (read stream t nil t))))))

    ;; consume any remaining clauses
    (let ((*read-suppress* t))
      (read-delimited-list #\) stream t))

    (values-list result)))


;;;;
;;;; stateful reads similar to CPP's #if/#elif/#else/#endif
;;;;

#|
  Unlike CPP's conditional inclusion, these don't have to be on line
  boundaries.

  It is an error for the input stream to end when a conditional block
  is suppressing READ; but it is acceptable (if unclean) for it to
  end when READ is not suppressed.

  e.g.
    "#(read-block-if nil)  ... EOF" signals an error
    but
    "#(read-block-if t) ... EOF" succeeds

  Each block is treated like a COND; at most one section is read.
|#

(defvar *read-block-stack* nil
  "track nesting/state between the read-block-* macros
Push t when the current block has allowed a read or the block was started
with *read-suppress*; nil otherwise.")
(defvar *read-block-suppress* nil
  "set this lexically when suppressing read")

;; Use a condition instead of setting a sentinal since conditions
;; can break through intervening reads.
(define-condition stop-suppressing
    (error) ; its an implementation error for this not to be caught
  () ; no slots
  (:documentation "Internal signal for read-block-*"))

(defun %read-block-else-if (stream subchar infix-param
                            test)
  (declare (ignore subchar))
  "Implementation detail"
  
  (when infix-param
    (error 'reader-error :stream stream))
  (when (read-delimited-list #\) stream t)
    (error 'reader-error :stream stream))
  (unless *read-block-stack*
    (error "Must be called inside a conditional read block."))

  ;; only allow the first true clause to be read
  (when (car *read-block-stack*)
    (setf test nil))

  (if test
      ;; let the next section be read
      (progn
        ;; mark this block as having a passed test
        (setf (car *read-block-stack*) t)
        ;; stop any suppression
        (when *read-block-suppress*
          (signal 'stop-suppressing)))
      
      ;; supress and discard failed sections
      (unless *read-block-suppress*
        (tagbody
           (handler-bind ((stop-suppressing
                           (lambda (c)
                             (declare (ignore c))
                             (go _done-suppressing_))))
             (let ((*read-suppress* t)
                   (*read-block-suppress* t))
               ;; discard reads until another read-block-* changes state
               (do ()
                   (nil)
                 (read stream t nil t))))
         _done-suppressing_)))
  (values))
  
(defun read-block-if (stream subchar infix-param)
  "#?(read-block-if test)
Like CPP's #if, begins a conditional read block.
Subsequent text will only be read if test is true."
  ;; start a new block; no tests have passed yet
  (push *read-suppress* *read-block-stack*)
  (let ((test (eval (read stream t nil t))))
    (%read-block-else-if stream subchar infix-param
                         test)))

(defun read-block-else-if (stream subchar infix-param)
  "#?(read-block-else-if test)
Like CPP's #elif, subsequent text will only be read if all previous
tests failed and this test is true.
Must be called inside a conditional read block."
  (let ((test (eval
               (let ((*read-suppress* (car *read-block-stack*)))
                 (read stream t nil t)))))
    (%read-block-else-if stream subchar infix-param
                         test)))

(defun read-block-else (stream subchar infix-param)
  "#?(read-block-else)
Like CPP's #else, subsequent text will only be read if all previous
tests failed.
Must be called inside a conditional read block."
  (%read-block-else-if stream subchar infix-param
                       t))

(defun read-block-end (stream subchar infix-param)
  (declare (ignore subchar))
  "#?(read-block-end)
Like CPP's #endif, ends a conditional read block.
Must be called inside a conditional read block."
  (when infix-param
    (error 'reader-error :stream stream))
  (when (read-delimited-list #\) stream t)
    (error 'reader-error :stream stream))
  (unless *read-block-stack*
    (error "Must be called inside a conditional read block."))
  (pop *read-block-stack*)
  ;; stop any suppression
  (when *read-block-suppress*
    (signal 'stop-suppressing))
  (values))


;;;;
;;;; lexical read manipulation
;;;;

(defun read-let (stream subchar infix-param)
  (declare (ignore subchar))
  "#?(read-let bindings form)
Lexically set BINDINGS while reading FORM.
Especially useful for things like *read-base* and other things in CLHS 2.1.2."
  (when infix-param
    (error 'reader-error :stream stream))
  (let ((symbols nil)
        (values nil)
        (bindings (read stream t nil t))
        (result nil))
    (cond
      ((null bindings)
       ;; nothing to do
       )
      ((listp bindings)
       (dolist (binding bindings)
         (cond ((symbolp binding)
                (push binding symbols)
                (push nil values))
               ((listp binding)
                (destructuring-bind (sym val) binding
                  (push sym symbols)
                  (push val values)))
               (t (error 'reader-error :stream stream)))))
      (t (error 'reader-error :stream stream)))

    (progv (reverse symbols) (reverse values)
      (setf result (multiple-value-list (read stream t nil t))))

    ;; check there isn't any junk after the form
    ;; and flush the #\) at the same time
    (when (read-delimited-list #\) stream t)
      (error 'reader-error :stream stream))

    (values-list result)))


;;;;
;;;; here documents
;;;;

(defconstant +whitespace+
  (if (boundp '+whitespace+)
      (symbol-value '+whitespace+)
      #(#\tab #\newline #\linefeed #\page #\return #\space))
  "whitespace chars from CLHS 2.1.4")

;; Idea: use the infix-param to trim N characters from the front of each line
;; (e.g. leading whitespace)
(defun read-here-string (stream subchar infix-param)
  (declare (ignore subchar))
  "\"#?(read-here delimiter) [text treated as comments]
 arbitrary text or (lisp expressions)
 delimiter\"
 Inspired by shell scripts (e.g. Bash).
 Read lines from the current file until a line containing only DELIMITER
 (with no leading or trailing whitespace) is found.  All lines before that
 point are then returned as a string.  No READ or EVAL is performed on the
 string.
 The delimiter cannot contain #\)."
  (when infix-param
    (error 'reader-error :stream stream))
  (let* ((delimiter (read-line stream t nil t))
         (start 0)
         (end (position #\) delimiter))
         (result nil))
    (do ()
        ((not (find (char delimiter start) +whitespace+)))
      (incf start))
    (unless end
      (error 'reader-error :stream stream))
    (setf delimiter (subseq delimiter start end))
    (do ((line (read-line stream t nil t)
               (read-line stream t nil t)))
        ((string= line delimiter))
      (push line result))
    (format nil "~{~A~#[~:;~%~]~}" (reverse result))))
